/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <errno.h>
#include <string.h>
#include <time.h>

#include <doca_buf.h>
#include <doca_buf_inventory.h>
#include <doca_dev.h>
#include <doca_error.h>
#include <doca_log.h>
#include <doca_mmap.h>
#include <doca_regex.h>
#include <doca_regex_mempool.h>

#include <utils.h>

DOCA_LOG_REGISTER(REGEX_SCAN);

#define NB_CHUNKS 6
#define SLEEP_IN_NANOS (10 * 1000) /* Sample the job every 10 microseconds  */

struct regex_scan_ctx {
	char *data_buffer;		       /* Data buffer */
	size_t data_buffer_len;		       /* Length of data buffer */
	char *rules_buffer;		       /* Rules buffer */
	size_t rules_buffer_len;	       /* Length of rules buffer */
	int qp_id;			       /* QP index */
	int chunk_len;			       /* size of chunk to send to regex */
	uint16_t nb_qp;			       /* Number of QPs to use */
	char *pci_address;		       /* RegEx PCI address to use */
	struct doca_buf *buf;		       /* active job buffer */
	struct doca_buf_inventory *buf_inv;    /* Pool of doca_buf objects */
	struct doca_dev *dev;		       /* DOCA device */
	struct doca_mmap *mmap;		       /* DOCA Memory orchestration */
	struct doca_regex *doca_regex;	       /* DOCA RegEx interface */
};

/*
 * Printing the RegEx results
 */
static void
regex_scan_report_results(struct regex_scan_ctx *regex_cfg, struct doca_regex_job_response *job_response, int chunk_len)
{
	int offset;
	struct doca_regex_match *ptr;

	if (job_response->num_matches == 0)
		return;
	ptr = job_response->matches;
	/* Match start is relative to the whole file data and not the current chunk */
	offset = job_response->id * chunk_len;
	while (ptr != NULL) {
		DOCA_LOG_INFO("date rule id: %d", ptr->rule_id);
		regex_cfg->data_buffer[ptr->match_start + offset + ptr->length] = '\0';
		DOCA_LOG_INFO("date value: %*s", ptr->length,
			      (char *)(regex_cfg->data_buffer + offset + ptr->match_start));
		struct doca_regex_match *const to_release_match = ptr;

		ptr = ptr->next;
		doca_regex_mempool_obj_put(job_response->matches_mempool, to_release_match);
	}
}

/*
 * Initialize DOCA RegEx resources
 * regex_cfg: RegEx configuration struct
 * Init DOCA RegEx according to regex_cfg configuration struct fields
 */
static int
regex_scan_init(struct regex_scan_ctx *regex_cfg)
{
	int ret;
	doca_error_t result;
	const int mempool_size = 8;
	struct doca_pci_bdf pci_bdf;

	/* find doca_dev */
	if (parse_pci_addr(regex_cfg->pci_address, &pci_bdf) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Invalid PCI address: \"%s\"", regex_cfg->pci_address);
		return -EINVAL;
	}
	if (open_doca_device_with_property(DOCA_DEVINFO_PROPERTY_PCI_ADDR, (uint8_t *)&pci_bdf, sizeof(pci_bdf),
			&regex_cfg->dev, NULL) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("No device matching PCI address found.");
		return -EINVAL;
	}

	/* Create a doca reg instance */
	regex_cfg->doca_regex = doca_regex_create();
	if (regex_cfg->doca_regex == NULL) {
		DOCA_LOG_ERR("Unable to create RegEx device.");
		return -ENOMEM;
	}

	/* Set the regex device as the main HW accelerator */
	ret = doca_regex_dev_add(regex_cfg->doca_regex, regex_cfg->dev);
	if (ret < 0) {
		DOCA_LOG_ERR("Unable to set RegEx device. [%s]", strerror(abs(ret)));
		return ret;
	}

	/* Configure number of QPs to use. */
	ret = doca_regex_num_qps_set(regex_cfg->doca_regex, regex_cfg->nb_qp);
	if (ret < 0) {
		DOCA_LOG_ERR("Unable to configure %d QPs. [%s]", regex_cfg->nb_qp, strerror(abs(ret)));
		return ret;
	}

	/* Size QP-0 memory pool, this app uses only 1 QP. */
	result = doca_regex_property_matches_memory_pool_size_set(regex_cfg->doca_regex, mempool_size);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable set matches mempool size. Reason: %s", doca_get_error_string(result));
		return -EINVAL;
	}

	/* Load compiled rules into the RegEx */
	result = doca_regex_property_hardware_binary_rules_set(regex_cfg->doca_regex, (uint8_t *)regex_cfg->rules_buffer,
			regex_cfg->rules_buffer_len);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to program rules file. Reason: %s", doca_get_error_string(result));
		return -ENOMEM;
	}

	/* Start doca RegEx */
	ret = doca_regex_start(regex_cfg->doca_regex);
	if (ret < 0) {
		DOCA_LOG_ERR("Unable to start doca RegEx. [%s]", strerror(abs(ret)));
		return ret;
	}

	result = doca_buf_inventory_create("inv", 1, DOCA_BUF_EXTENSION_NONE, &regex_cfg->buf_inv);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to create buffer inventory. Reason: %s", doca_get_error_string(result));
		return -ENOMEM;
	}

	result = doca_buf_inventory_start(regex_cfg->buf_inv);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to start buffer inventory. Reason: %s", doca_get_error_string(result));
		return -ENOMEM;
	}

	result = doca_mmap_create("mmap", &regex_cfg->mmap);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to create memory map. Reason: %s", doca_get_error_string(result));
		return -ENOMEM;
	}

	result = doca_mmap_start(regex_cfg->mmap);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to start memory map. Reason: %s", doca_get_error_string(result));
		return -ENOMEM;
	}

	result = doca_mmap_dev_add(regex_cfg->mmap, regex_cfg->dev);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to add device to memory map. Reason: %s", doca_get_error_string(result));
		return -ENOMEM;
	}

	result = doca_mmap_populate(regex_cfg->mmap, regex_cfg->data_buffer, regex_cfg->data_buffer_len, 1024 * 4, NULL,
				    NULL);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to add memory region to memory map. Reason: %s", doca_get_error_string(result));
		return -ENOMEM;
	}

	return 0;
}

/*
 * Enqueue job to DOCA RegEx qp
 * regex_cfg: regex_scan_ctx configuration struct
 * job_request: RegEx job request, already initialized with first chunk.
 * remaining_bytes: the remaining bytes to send all jobs (chunks).
 */
static uint32_t
regex_scan_enq_job(struct regex_scan_ctx *regex_cfg, struct doca_regex_job_request *job_request,
		   uint32_t *remaining_bytes)
{
	int enq_result;
	uint32_t nb_enqueued = 0;
	uint32_t nb_free = 0;

	doca_buf_inventory_num_free_elements_read(regex_cfg->buf_inv, &nb_free);

	if (*remaining_bytes != 0 && nb_free != 0) {
		struct doca_buf *buf;
		uint32_t const job_size =
			regex_cfg->chunk_len < *remaining_bytes ? regex_cfg->chunk_len : *remaining_bytes;
		uint32_t const read_offset = regex_cfg->data_buffer_len - *remaining_bytes;

		if (doca_buf_inventory_buf_by_addr(regex_cfg->buf_inv, regex_cfg->mmap,
						   regex_cfg->data_buffer + read_offset, job_size,
						   &buf) != DOCA_SUCCESS)
			return nb_enqueued;

		regex_cfg->buf = buf;
		job_request->buffer = buf;
		enq_result = doca_regex_enqueue(regex_cfg->doca_regex, regex_cfg->qp_id, job_request, false);
		if (enq_result == 0) {
			doca_buf_refcount_rm(buf, NULL);
			return nb_enqueued; /* QP is full, try to dequeue. */
		}
		if (enq_result < 0)
			APP_EXIT("Unable to enqueue job. [%s]", strerror(abs(enq_result)));
		*remaining_bytes -= job_size; /* Update remaining bytes to scan. */
		nb_enqueued++;
		--nb_free;

		/* Prepare next chunk. */
		job_request->id++;
	}

	return nb_enqueued;
}

static uint32_t
regex_scan_deq_job(struct regex_scan_ctx *regex_cfg, int chunk_len)
{
	int ret;
	uint32_t nb_dequeued = 0;
	struct doca_regex_job_response job_responses[1];
	struct timespec ts;

	do {
		ret = doca_regex_dequeue(regex_cfg->doca_regex, regex_cfg->qp_id, &job_responses[0], 1);
		if (ret < 0)
			APP_EXIT("Failed to dequeue results. [%s]", strerror(abs(ret)));
		else if (ret == 0) {
			/* Wait for the job to complete */
			ts.tv_sec = 0;
			ts.tv_nsec = SLEEP_IN_NANOS;
			nanosleep(&ts, &ts);
		} else {
			/* release the buffer back into the pool so it can be re-used */
			doca_buf_refcount_rm(regex_cfg->buf, NULL);
			regex_scan_report_results(regex_cfg, &job_responses[0], chunk_len);
			++nb_dequeued;
		}

	} while (ret != 0);

	return nb_dequeued;
}

/*
 * RegEx scan cleanup, destroy all DOCA RegEx resources
 */
static void
regex_scan_destroy(struct regex_scan_ctx *regex_cfg)
{
	doca_regex_stop(regex_cfg->doca_regex);
	doca_regex_destroy(regex_cfg->doca_regex);
	regex_cfg->doca_regex = NULL;

	if (regex_cfg->mmap != NULL) {
		doca_mmap_dev_rm(regex_cfg->mmap, regex_cfg->dev);
		doca_mmap_stop(regex_cfg->mmap);
		doca_mmap_destroy(regex_cfg->mmap);
		regex_cfg->mmap = NULL;
	}

	if (regex_cfg->buf_inv != NULL) {
		doca_buf_inventory_stop(regex_cfg->buf_inv);
		doca_buf_inventory_destroy(regex_cfg->buf_inv);
		regex_cfg->buf_inv = NULL;
	}

	if (regex_cfg->dev != NULL) {
		doca_dev_close(regex_cfg->dev);
		regex_cfg->dev = NULL;
	}
}

/*
 * Run DOCA regex sample
 * data_buffer: User data used to find the matches
 * data_buffer_len: data_buffer length
 * pci_addr: pci address for HW RegEx device
 * rules_buffer: Rules data (compiled rules(rof2.binary))
 * rules_buffer_len: rules_buffer length
 */
int
regex_scan(char *data_buffer, size_t data_buffer_len, char *pci_addr, char *rules_buffer, size_t rules_buffer_len)
{
	if (data_buffer == NULL || data_buffer_len == 0 || pci_addr == NULL || rules_buffer == NULL ||
	    rules_buffer_len == 0)
		return -EINVAL;

	int ret;
	uint32_t remaining_bytes, nb_dequeued = 0, nb_enqueued = 0;
	const int nb_chunks = NB_CHUNKS;
	struct regex_scan_ctx rgx_cfg = {0};
	struct doca_regex_job_request job_request = {0};

	/* Set DOCA RegEx configuration fields in regex_cfg according to our sample */
	rgx_cfg.nb_qp = 1;
	rgx_cfg.qp_id = 0;
	rgx_cfg.data_buffer = data_buffer;
	rgx_cfg.data_buffer_len = data_buffer_len;
	rgx_cfg.pci_address = pci_addr;
	rgx_cfg.rules_buffer = rules_buffer;
	rgx_cfg.rules_buffer_len = rules_buffer_len;
	rgx_cfg.chunk_len = (rgx_cfg.data_buffer_len < nb_chunks) ? rgx_cfg.data_buffer_len
								  : 1 + (rgx_cfg.data_buffer_len / nb_chunks);

	/* Init DOCA RegEx */
	ret = regex_scan_init(&rgx_cfg);
	if (ret < 0)
		return ret;

	job_request.rule_group_ids[0] = 1;

	remaining_bytes = data_buffer_len;

	/* The main loop, enqueues jobs (chunks) and dequeues for results. */
	do {
		nb_enqueued += regex_scan_enq_job(&rgx_cfg, &job_request, &remaining_bytes);
		nb_dequeued += regex_scan_deq_job(&rgx_cfg, rgx_cfg.chunk_len);
	} while (remaining_bytes > 0 || nb_dequeued != nb_enqueued);

	/* RegEx scan recognition cleanup */
	regex_scan_destroy(&rgx_cfg);
	return 0;
}
