/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */


#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <doca_log.h>
#include <doca_telemetry.h>

#define NB_EXAMPLE_STRINGS 5		/* Amount of example strings */
#define MAX_EXAMPLE_STRING_SIZE 256	/* Indicates the max length of string */
#define SINGLE_FIELD_VALUE 1		/* Indicates the field contains one value */

DOCA_LOG_REGISTER(TELEMETRY);

static char *example_strings[NB_EXAMPLE_STRINGS] = {
	"example_str_1",
	"example_str_2",
	"example_str_3",
	"example_str_4",
	"example_str_5"
};

/* Event struct from which report will be serialized */
struct test_event_type {
	doca_telemetry_timestamp_t  timestamp;
	int32_t                     event_number;
	int32_t                     iter_num;
	uint64_t                    string_number;
	char                        example_string[MAX_EXAMPLE_STRING_SIZE];
} __attribute__((packed));


/*
 * This function fills up event buffer with the example string of specified number.
 * It also saves number of iteration, number of string and overall number of events.
 */
static doca_error_t
prepare_example_event(struct test_event_type *ev1, int iter_num, int string_number)
{
	static int collected_example_events_count; /* Initalized to 0 by default, only in first call to function */
	doca_telemetry_timestamp_t timestamp;
	doca_error_t ret = DOCA_SUCCESS;

	ret = doca_telemetry_timestamp_get(&timestamp);
	if (ret != DOCA_SUCCESS)
		return ret;

	ev1->timestamp = timestamp;
	ev1->event_number  = collected_example_events_count++;
	ev1->iter_num      = iter_num;
	ev1->string_number = string_number;
	if (strlen(example_strings[string_number]) >= MAX_EXAMPLE_STRING_SIZE)
		return DOCA_ERROR_NO_MEMORY;
	strcpy(ev1->example_string, example_strings[string_number]);
	return ret;
}

static doca_error_t
telemetry_register_fields(struct doca_telemetry_type *type)
{
	doca_error_t ret;
	struct doca_telemetry_field *field;
	const int nb_fields = 5;
	int idx = 0;
	struct {
		const char *name;
		const char *desc;
		const char *type_name;
		uint16_t len;
	} fields_info[] = {
		{"iter_num", "Iteration number", DOCA_TELEMETRY_FIELD_TYPE_INT32, SINGLE_FIELD_VALUE},
		{"timestamp", "Event timestamp", DOCA_TELEMETRY_FIELD_TYPE_TIMESTAMP, SINGLE_FIELD_VALUE},
		{"event_number", "Event number", DOCA_TELEMETRY_FIELD_TYPE_INT32, SINGLE_FIELD_VALUE},
		{"string_number", "String number", DOCA_TELEMETRY_FIELD_TYPE_UINT64, SINGLE_FIELD_VALUE},
		{"example_string", "String example", DOCA_TELEMETRY_FIELD_TYPE_CHAR, MAX_EXAMPLE_STRING_SIZE},
	};

	for (idx = 0; idx < nb_fields; idx++) {
		ret = doca_telemetry_field_create(&field);
		if (ret != DOCA_SUCCESS)
			return ret;

		doca_telemetry_field_set_name(field, fields_info[idx].name);
		doca_telemetry_field_set_description(field, fields_info[idx].desc);
		doca_telemetry_field_set_type_name(field, fields_info[idx].type_name);
		doca_telemetry_field_set_array_length(field, fields_info[idx].len);

		ret = doca_telemetry_type_add_field(type, field);
		if (ret != DOCA_SUCCESS) {
			doca_telemetry_type_destroy(type);
			return ret;
		}
	}

	return ret;
}

int
telemetry_export()
{
	bool file_write_enable = true;		/* Enables writing to local machine as file */
	bool ipc_enabled = true;		/* Enables sending to DTS through ipc sockets */
	int repetition = 10;			/* Repetition amount of exporting telemetry */
	doca_error_t ret;
	int k = 0;
	int i = 0;
	struct doca_telemetry_schema *doca_schema = NULL;
	struct doca_telemetry_source *source = NULL;
	struct test_event_type test_event;
	doca_telemetry_type_index_t example_index;

	/* Event type for schema. Should be consistent with event struct. */
	struct doca_telemetry_type *example_fields;

	/* =============================== SCHEMA =============================== */

	/* Init DOCA schema */
	ret = doca_telemetry_schema_init("example_doca_schema_name", &doca_schema);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("cannot init doca schema");
		return 1;
	}

	/* ======================== SCHEMA ATTRIBUTES =========================== */

	/* Set buffer size in bytes to fit 5 example events. By default it is set to 60K.
	 * Data root should be set to keep data schemas and binary data if file_write
	 * is enabled.
	 */
	doca_telemetry_schema_set_buffer_size(doca_schema, sizeof(test_event) * 5);

	ret = doca_telemetry_type_create(&example_fields);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("cannot create type");
		return 1;
	}

	ret = telemetry_register_fields(example_fields);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("cannot register fields");
		doca_telemetry_type_destroy(example_fields);
		return 1;
	}

	/* Enable file write during the app development.
	 * Check written files under data root to make sure that data format is correct.
	 * Default max_file_size is 1 Mb, default max_file_age is 1 hour.
	 */
	if (file_write_enable)
		doca_telemetry_schema_set_file_write_enabled(doca_schema);

	/* If ipc is enabled, doca telemetry will try to find Telemetry Service socket
	 * under ipc_sockets_dir. IPC is disabled by default.
	 * Optionally change parameters for IPC connection/reconnection tries
	 * and IPC socket timeout. Default values are 100 msec, 3 tries, and 500 ms accordingly.
	 * see doca_telemetry_schema_set_ipc_* functions for details.
	 */
	if (ipc_enabled)
		doca_telemetry_schema_set_ipc_enabled(doca_schema);

	/* Add schema types */
	ret = doca_telemetry_schema_add_type(doca_schema, "example_event", example_fields, &example_index);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("cannot add type to doca_schema!");
		goto err_schema;
	}
	/* "apply" schema */
	ret = doca_telemetry_schema_start(doca_schema);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("cannot start doca_schema!");
		goto err_schema;
	}

	/* =========================== SCHEMA END =============================== */


	/* ======================Create Telemetry source ======================== */

	/* Create DOCA Telemetry Source context from DOCA schema */
	ret = doca_telemetry_source_create(doca_schema, &source);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("cannot create doca_source!");
		goto err_schema;
	}

	doca_telemetry_source_set_id(source, "source_1");
	doca_telemetry_source_set_tag(source, "source_1_tag");

	/* Start source to apply attributes and start services */
	ret = doca_telemetry_source_start(source);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("cannot start doca_source!");
		goto err_source;
	}
	/* Create more DOCA Sources if needed. */

	/* Prepare events and report them via DOCA Telemetry */
	for (k = 0; k < repetition; k++) {
		for (i = 0; i < NB_EXAMPLE_STRINGS; i++) {
			DOCA_LOG_INFO("progressing: k=%d \t i=%d", k, i);
			if (prepare_example_event(&test_event, k, i) != DOCA_SUCCESS) {
				DOCA_LOG_ERR("failed to create event");
				goto err_source;
			}
			if (doca_telemetry_source_report(source, example_index,
							 &test_event, 1) != DOCA_SUCCESS) {
				DOCA_LOG_ERR("cannot report to doca_source!");
				goto err_source;
			}
		}
		if (k % 2 == 0) {
			/*
			 * Optionally force DOCA source buffer to flush.
			 * Handy for bursty events or specific event types.
			 */
			ret = doca_telemetry_source_flush(source);
			if (ret != DOCA_SUCCESS) {
				DOCA_LOG_ERR("cannot flush doca_source!");
				goto err_source;
			}
		}
	}

	/* Destroy all DOCA sources and DOCA schema to clean up */
	doca_telemetry_source_destroy(source);
	doca_telemetry_schema_destroy(doca_schema);

	return 0;
err_source:
	doca_telemetry_source_destroy(source);
err_schema:
	doca_telemetry_schema_destroy(doca_schema);
	return 1;
}
